use backon::BlockingRetryable;
use backon::ExponentialBuilder;
use clap::ValueEnum;
use log::{debug, error, info};
use std::{
    net::{IpAddr, SocketAddr, ToSocketAddrs, UdpSocket},
    time::Duration,
};
use tailstun::StunMessage;

#[derive(Debug, Clone, ValueEnum)]
enum OutputFormat {
    Text,
    Json,
    Yaml,
}

impl OutputFormat {
    fn format_stun(&self, msg: &StunMessage) -> String {
        match self {
            OutputFormat::Text => format!("{}", msg),
            OutputFormat::Json => serde_json::to_string_pretty(msg).unwrap(),
            OutputFormat::Yaml => serde_yaml::to_string(msg).unwrap(),
        }
    }
    fn format_address(&self, a: &SocketAddr) -> String {
        let a = match a {
            SocketAddr::V4(_) => a.ip(),
            SocketAddr::V6(v6) => {
                if let Some(v4) = v6.ip().to_ipv4_mapped() {
                    IpAddr::V4(v4)
                } else {
                    a.ip()
                }
            }
        };
        match self {
            OutputFormat::Text => format!("{}", a),
            OutputFormat::Json => serde_json::to_string_pretty(&a).unwrap(),
            OutputFormat::Yaml => serde_yaml::to_string(&a).unwrap(),
        }
    }
}

#[derive(clap::Parser)]
#[command(version, about, long_about = None)]
#[command(about = "Test a Tailscale derp node's stun service")]
struct Cli {
    host: String,
    #[clap(default_value = "3478")]
    port: u16,
    #[clap(
        short = '4',
        conflicts_with = "v6_only",
        default_value = "false",
        help = "Only use IPv4"
    )]
    v4_only: bool,
    #[clap(
        short = '6',
        conflicts_with = "v4_only",
        default_value = "false",
        help = "Only use IPv6"
    )]
    v6_only: bool,
    #[clap(short, long, default_value = "text")]
    format: OutputFormat,
    #[clap(
        short,
        long,
        default_value = "false",
        help = "Only output the first mapped address & convert IPv6-mapped to IPv4"
    )]
    address_only: bool,
    #[clap(
        short,
        long,
        default_value = "5.0",
        value_parser = parse_duration,
        help = "Timeout in seconds"
    )]
    timeout: Duration,
    #[clap(short, long, default_value = "0", help = "Number of retries")]
    retries: usize,
    #[command(flatten)]
    verbose: clap_verbosity_flag::Verbosity,
}

fn parse_duration(s: &str) -> Result<Duration, std::num::ParseFloatError> {
    let secs = s.parse()?;
    Ok(Duration::from_secs_f64(secs))
}

fn stun_query(
    target: &SocketAddr,
    timeout: Duration,
    retries: usize,
) -> Result<StunMessage, std::io::Error> {
    let socket = UdpSocket::bind("[::]:0").expect("Unable to bind a UDP socket");
    socket
        .connect(target)
        .expect("Unable to connect to the destination");
    socket
        .set_read_timeout(Some(timeout))
        .expect("Unable to set read timeout");
    debug!(
        "Connected UDP socket to {:?} from {:?}",
        socket.peer_addr(),
        socket.local_addr()
    );

    debug!("Building request packet");
    let req = tailstun::rand_request();
    let backoff = ExponentialBuilder::default().with_max_times(retries);

    debug!("request {:?}", &req);

    let mut buf = [0u8; 1500];
    let fetch = || {
        info!("Sending STUN request to {target} with timeout {timeout:?}");
        socket.send(&req)?;
        socket.recv(&mut buf)
    };
    let response = fetch
        .retry(backoff)
        .when(|r| r.kind() == std::io::ErrorKind::WouldBlock)
        .notify(|_err, dur| info!("retrying after {:?}", dur))
        .call();

    match response {
        Ok(received) => {
            let buf = &buf[..received];
            debug!("Received response: {:?}", buf);

            let msg = StunMessage::parse(buf).unwrap();
            info!("Parsed message from {}:", socket.peer_addr().unwrap());
            Ok(msg)
        }
        Err(e) => Err(e),
    }
}

fn main() {
    let cli = <Cli as clap::Parser>::parse();
    env_logger::Builder::new()
        .filter_level(cli.verbose.log_level_filter())
        .init();

    let dest = (cli.host.as_str(), cli.port)
        .to_socket_addrs()
        .expect("Unable to resolve host")
        .find(|a| {
            if cli.v4_only {
                a.is_ipv4()
            } else if cli.v6_only {
                a.is_ipv6()
            } else {
                true
            }
        })
        .expect("No address found for host");

    let msg = match stun_query(&dest, cli.timeout, cli.retries) {
        Ok(msg) => msg,
        Err(e) => {
            if e.kind() == std::io::ErrorKind::WouldBlock {
                error!("Timed out after all retries used, giving up.");
            } else {
                error!("Error: {:?}", e);
            }
            std::process::exit(1);
        }
    };
    if cli.address_only {
        match msg.attributes.mapped_address() {
            Some(addr) => println!("{}", cli.format.format_address(addr)),
            None => {
                // No mapped address
                std::process::exit(1);
            }
        }
    } else {
        println!("{}", cli.format.format_stun(&msg));
    }
}
